__title__ = "WakaQ"
__description__ = "Background task queue for Python backed by Redis, a minimal Celery."
__url__ = "https://github.com/wakatime/wakaq"
__version_info__ = ("2", "1", "9")
__version__ = ".".join(__version_info__)
__author__ = "Alan Hamlett"
__author_email__ = "alan.hamlett@gmail.com"
__license__ = "BSD"
__copyright__ = "Copyright 2022 Alan Hamlett"
